/**
 * Another Ivy java library API tester: requests
 *
 * @author  Yannick Jestin <mailto:jestin@cena.fr>
 *
 * (c) CENA
 *
 * usage: java Request
 *
 * Changelog
 *   1.2.8 : first release
 *
 */
import fr.dgac.ivy.*;

class Request {

  public static void main(String[] args) throws IvyException {
    String domain=Ivy.getDomainArgs("RequestTest",args);
    new Request(domain);
  }

  private Ivy bus;
  static int serial = 0;
  private static final int NBCOMP = 5;

  public Request(String domain) throws IvyException {
    bus = new Ivy("RequestTest","RequestTest ready", null);
    String id=bus.getWBUId();
    bus.bindMsgOnce("^result id="+id+" value=([0-9]+)",new IvyMessageListener() {
      public void receive(IvyClient ic,String[] args) {
	System.out.println("result received: "+args[0]);
      }
    });
    bus.start(domain);
    System.out.println("launching "+NBCOMP+" Computers");
    for (int i=0;i<NBCOMP;i++) new Computer(domain);
    System.out.println("waiting for them");
    System.out.println("going to sleep 2 seconds");
    try { Thread.sleep(2*1000); } catch (InterruptedException ie) { }
    for (int i=0;i<NBCOMP;i++) bus.waitForClient("Computer"+i,0); // be sure there are all here
    bus.sendMsg("computesum id="+id+" a=3 b=4");
    System.out.println("going to sleep 2 seconds");
    try { Thread.sleep(2*1000); } catch (InterruptedException ie) { }
    System.out.println("awaking and leaving the bus");
    bus.stop();
  }

  private class Computer implements IvyMessageListener {
    Ivy bus;
    String name;
    public Computer(String domain) throws IvyException {
      name = "Computer"+serial++;
      bus = new Ivy(name,name+" ready",null);
      bus.bindMsg("^computesum id=([^ ]*) a=([0-9]*) b=([0-9]*)",this);
      bus.start(domain);
    }
    public void receive(IvyClient ic,String[] args) {
      String id=args[0];
      int a=Integer.parseInt(args[1]);
      int b=Integer.parseInt(args[2]);
      System.out.println(name+" sending result, id: "+id+", a:"+a+", b:"+b);
      try { bus.sendMsg("result id="+id+" value="+(a+b));}
      catch (IvyException ie) { ie.printStackTrace(); }
      bus.stop();
    }
  }

}
